from __future__ import annotations

import re
import sys
from collections import deque
from typing import Optional, Union

from sec_downloader.types import FilingMetadata, RequestedFilings, Ticker
from sec_edgar_downloader import Downloader
from sec_edgar_downloader._constants import (
    AMENDS_SUFFIX,
    CIK_LENGTH,
    SUBMISSION_FILE_FORMAT,
    SUPPORTED_FORMS,
    URL_SUBMISSIONS,
)
from sec_edgar_downloader._orchestrator import get_to_download
from sec_edgar_downloader._sec_gateway import get_list_of_available_filings
from sec_edgar_downloader._utils import validate_and_convert_ticker_or_cik

accession_number_re = re.compile(r"^\d{10}-\d{2}-\d{6}$")


def get_filing_metadata(
    *,
    ticker_or_cik: str,
    accession_number: str,
    user_agent: str,
    ticker_to_cik_mapping: dict[str, str],
) -> FilingMetadata:
    if len(accession_number) == 18:
        accession_number = (
            f"{accession_number[:10]}-{accession_number[10:12]}-{accession_number[12:]}"
        )
    if not accession_number_re.match(accession_number):
        raise ValueError(f"Invalid Accession Number: {accession_number}")
    cik = validate_and_convert_ticker_or_cik(ticker_or_cik, ticker_to_cik_mapping)
    result = _get_metadatas(
        cik=cik,
        user_agent=user_agent,
        limit=1,
        accession_number=accession_number,
    )
    if len(result) == 0:
        raise ValueError(f"Could not find filing for {accession_number}")
    assert len(result) == 1
    return result[0]


def get_latest_filings_metadata(
    *,
    requested: RequestedFilings,
    user_agent: str,
    ticker_to_cik_mapping: dict[str, str],
) -> list[FilingMetadata]:
    cik = validate_and_convert_ticker_or_cik(
        requested.ticker_or_cik, ticker_to_cik_mapping
    )

    if requested.limit is None:
        # If amount is not specified, obtain all available filings.
        # We simply need a large number to denote this and the loop
        # responsible for fetching the URLs will break appropriately.
        limit = sys.maxsize
    else:
        limit = int(requested.limit)
        if limit < 1:
            raise ValueError("Invalid amount. Please enter a number greater than 1.")

    if requested.form_type not in SUPPORTED_FORMS:
        form_options = ", ".join(Downloader.supported_forms)
        raise ValueError(
            f"{requested.form_type!r} forms are not supported. "
            f"Please choose from the following: {form_options}."
        )

    return _get_metadatas(
        cik=cik,
        user_agent=user_agent,
        limit=limit,
        ticker_or_cik=requested.ticker_or_cik,
        form_type=requested.form_type,
    )


def _get_metadatas(
    *,
    cik: str,
    user_agent: str,
    limit: int,
    ticker_or_cik: Optional[str] = None,
    accession_number: Optional[str] = None,
    form_type: Optional[str] = None,
) -> list[FilingMetadata]:
    assert (
        ticker_or_cik and form_type
    ) or accession_number, (
        "Either ticker or CIK and form type or accession number must be provided"
    )

    submissions_uri = URL_SUBMISSIONS.format(
        submission=SUBMISSION_FILE_FORMAT.format(cik=cik)
    )

    additional_submissions = None
    found_metadatas: list[FilingMetadata] = []
    fetched_count = 0
    company_tickers = None
    company_cik = None
    company_name = None
    while fetched_count < limit:
        resp_json = get_list_of_available_filings(submissions_uri, user_agent)
        # First API response is different from further API responses
        if additional_submissions is None:
            filings_json = resp_json["filings"]["recent"]
            additional_submissions = deque(resp_json["filings"]["files"])
            company_tickers = [
                Ticker(symbol=ticker, exchange=exchange)
                for ticker, exchange in zip(
                    resp_json["tickers"], resp_json["exchanges"]
                )
            ]
            company_name = resp_json["name"]
            company_cik = str(resp_json["cik"]).zfill(CIK_LENGTH)
        # On second page or more of API response (for companies with >1000 filings)
        else:
            filings_json = resp_json
        assert company_tickers is not None
        assert company_name is not None
        assert company_cik is not None

        accession_numbers = filings_json["accessionNumber"]
        primary_document_urls = filings_json["primaryDocument"]
        filing_dates = filings_json["filingDate"]
        report_dates = filings_json["reportDate"]
        primary_doc_descriptions = filings_json["primaryDocDescription"]
        items_list = filings_json["items"]
        form_types = filings_json["form"]

        for (
            this_accession_number,
            primary_doc_filename,
            filing_date,
            report_date,
            primary_doc_description,
            this_form_type,
            items,
        ) in zip(
            accession_numbers,
            primary_document_urls,
            filing_dates,
            report_dates,
            primary_doc_descriptions,
            form_types,
            items_list,
        ):
            is_amend = this_form_type.endswith(AMENDS_SUFFIX)
            this_form_type = this_form_type[:-2] if is_amend else this_form_type
            if (
                (form_type and form_type != this_form_type)
                or (accession_number and accession_number != this_accession_number)
                or is_amend
            ):
                continue

            td = get_to_download(cik, this_accession_number, primary_doc_filename)
            found_metadata = FilingMetadata(
                primary_doc_url=td.primary_doc_uri,
                accession_number=this_accession_number,
                tickers=company_tickers,
                company_name=company_name,
                filing_date=filing_date,
                report_date=report_date,
                primary_doc_description=primary_doc_description,
                items=items,
                form_type=this_form_type,
                cik=company_cik,
            )
            found_metadatas.append(found_metadata)
            fetched_count += 1
            # We have reached the requested download limit, so break inner for loop
            # early and allow the outer while loop to break.
            if fetched_count == limit:
                break

        if len(additional_submissions) == 0:
            break

        next_page = additional_submissions.popleft()["name"]
        submissions_uri = URL_SUBMISSIONS.format(submission=next_page)

    requested_form = f" of type {form_type}" if form_type else ""
    error_context = f"{accession_number or ticker_or_cik}{requested_form}"
    if not found_metadatas:
        msg = f"Could not find any filings: {error_context}"
        raise ValueError(msg)
    if len(found_metadatas) > limit:
        msg = f"Found more than {limit} filings, actual count is {len(found_metadatas)}: {error_context}"
        raise OverflowError(msg)

    return found_metadatas



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/sec-downloader',
                'doc_host': 'https://Elijas.github.io',
                'git_url': 'https://github.com/Elijas/sec-downloader',
                'lib_path': 'sec_downloader'},
  'syms': { 'sec_downloader.core': {},
            'sec_downloader.download_storage': {},
            'sec_downloader.sec_edgar_downloader_fork': {},
            'sec_downloader.types': { 'sec_downloader.types.CompanyAndAccessionNumber': ( 'types.html#companyandaccessionnumber',
                                                                                          'sec_downloader/types.py'),
                                      'sec_downloader.types.CompanyAndAccessionNumber.from_string': ( 'types.html#companyandaccessionnumber.from_string',
                                                                                                      'sec_downloader/types.py'),
                                      'sec_downloader.types.FilingMetadata': ('types.html#filingmetadata', 'sec_downloader/types.py'),
                                      'sec_downloader.types.RequestedFilings': ('types.html#requestedfilings', 'sec_downloader/types.py'),
                                      'sec_downloader.types.RequestedFilings.from_string': ( 'types.html#requestedfilings.from_string',
                                                                                             'sec_downloader/types.py'),
                                      'sec_downloader.types.Ticker': ('types.html#ticker', 'sec_downloader/types.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_types.ipynb.

# %% auto 0
__all__ = ['Ticker', 'FilingMetadata', 'RequestedFilings', 'CompanyAndAccessionNumber']

# %% ../nbs/00_types.ipynb 1
import re
from dataclasses import dataclass
from typing import Optional

# %% ../nbs/00_types.ipynb 4
@dataclass
class Ticker:
    symbol: str
    exchange: str


@dataclass
class FilingMetadata:
    accession_number: str
    form_type: str
    primary_doc_url: str
    items: str
    primary_doc_description: str
    filing_date: str
    report_date: str
    cik: str
    company_name: str
    tickers: list[Ticker]

# %% ../nbs/00_types.ipynb 5
@dataclass
class RequestedFilings:
    ticker_or_cik: str
    form_type: str = "10-Q"
    limit: int = 1

    _REGEX_PATTERN = r"^(?:(\d+)/)?([^/]+)(?:/([^/]+))?$"

    @classmethod
    def from_string(cls, s: str):
        match = re.match(cls._REGEX_PATTERN, s.strip())
        if not match:
            raise ValueError(f"Invalid RequestedFilings string: {s}")
        limit_str, ticker_or_cik, form_type = match.groups()
        limit = int(limit_str) if limit_str else cls.limit
        form_type = form_type if form_type else cls.form_type
        form_type = form_type.upper()
        return cls(
            limit=limit,
            ticker_or_cik=ticker_or_cik,
            form_type=form_type,
        )

# %% ../nbs/00_types.ipynb 7
@dataclass
class CompanyAndAccessionNumber:
    ticker_or_cik: str
    accession_number: str

    _REGEX_PATTERN = re.compile(r"\b([A-Za-z0-9.]+)/(\d{10}-?\d{2}-?\d{6})\b")

    @classmethod
    def from_string(
        cls, s: str, *, must_match=False
    ) -> Optional["CompanyAndAccessionNumber"]:
        match = re.search(cls._REGEX_PATTERN, s.strip())
        if not match:
            if must_match:
                raise ValueError(f"Invalid RequestedFilings string: {s}")
            else:
                return None
        ticker_or_cik, accession_number = match.groups()
        return cls(
            ticker_or_cik=ticker_or_cik,
            accession_number=accession_number,
        )



import shutil
import tempfile
from glob import glob
from pathlib import Path
from typing import Optional

from sec_downloader.core import DEFAULT_FILTER_PATTERN, FileContent


class DownloadStorage:
    def __init__(self, *, filter_pattern: Optional[str] = None):
        self.glob_pattern = filter_pattern or DEFAULT_FILTER_PATTERN
        self.temp_dir = None
        self.file_contents = None

    def __enter__(self):
        self.temp_dir = tempfile.mkdtemp()
        return self.temp_dir

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._read_files()
        if self.temp_dir:
            shutil.rmtree(self.temp_dir)

    def _read_files(self):
        self.file_contents = []
        assert self.temp_dir is not None, "Temp dir should be set"
        glob_path = Path(self.temp_dir) / self.glob_pattern
        for filepath in glob(str(glob_path), recursive=True):
            path = Path(filepath)
            relative_path = path.relative_to(self.temp_dir)
            with open(filepath, "r", encoding="utf-8", errors="ignore") as f:
                content = f.read()
            self.file_contents.append(FileContent(relative_path, content))

    def get_file_contents(self):
        if self.file_contents is None:
            raise RuntimeError(
                "File contents are not available until the context is exited."
            )
        return self.file_contents



import shlex
from configparser import ConfigParser

import setuptools
from pkg_resources import parse_version

assert parse_version(setuptools.__version__) >= parse_version("36.2")

# note: all settings are in settings.ini; edit there, not here
config = ConfigParser(delimiters=["="])
config.read("settings.ini", encoding="utf-8")
cfg = config["DEFAULT"]

cfg_keys = "version description keywords author author_email".split()
expected = (
    cfg_keys
    + "lib_name user branch license status min_python audience language".split()
)
for o in expected:
    assert o in cfg, "missing expected setting: {}".format(o)
setup_cfg = {o: cfg[o] for o in cfg_keys}

licenses = {
    "apache2": (
        "Apache Software License 2.0",
        "OSI Approved :: Apache Software License",
    ),
    "mit": ("MIT License", "OSI Approved :: MIT License"),
    "gpl2": (
        "GNU General Public License v2",
        "OSI Approved :: GNU General Public License v2 (GPLv2)",
    ),
    "gpl3": (
        "GNU General Public License v3",
        "OSI Approved :: GNU General Public License v3 (GPLv3)",
    ),
    "bsd3": ("BSD License", "OSI Approved :: BSD License"),
}
statuses = [
    "1 - Planning",
    "2 - Pre-Alpha",
    "3 - Alpha",
    "4 - Beta",
    "5 - Production/Stable",
    "6 - Mature",
    "7 - Inactive",
]
py_versions = "3.6 3.7 3.8 3.9 3.10 3.11 3.12".split()

requirements = shlex.split(cfg.get("requirements", ""))
if cfg.get("pip_requirements"):
    requirements += shlex.split(cfg.get("pip_requirements", ""))
min_python = cfg["min_python"]
lic = licenses.get(cfg["license"].lower(), (cfg["license"], None))
dev_requirements = (cfg.get("dev_requirements") or "").split()

setuptools.setup(
    name=cfg["lib_name"],
    license=lic[0],
    classifiers=[
        "Development Status :: " + statuses[int(cfg["status"])],
        "Intended Audience :: " + cfg["audience"].title(),
        "Natural Language :: " + cfg["language"].title(),
    ]
    + [
        "Programming Language :: Python :: " + o
        for o in py_versions[py_versions.index(min_python) :]
    ]
    + (["License :: " + lic[1]] if lic[1] else []),
    url=cfg["git_url"],
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=requirements,
    extras_require={"dev": dev_requirements},
    dependency_links=cfg.get("dep_links", "").split(),
    python_requires=">=" + cfg["min_python"],
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    zip_safe=False,
    entry_points={
        "console_scripts": cfg.get("console_scripts", "").split(),
        "nbdev": [f'{cfg.get("lib_path")}={cfg.get("lib_path")}._modidx:d'],
    },
    **setup_cfg,
)



from collections import namedtuple
from typing import Optional, Union

from sec_downloader.sec_edgar_downloader_fork import (
    FilingMetadata,
    get_filing_metadata,
    get_latest_filings_metadata,
)
from sec_downloader.types import CompanyAndAccessionNumber, RequestedFilings
from sec_edgar_downloader import Downloader as SecEdgarDownloader
from sec_edgar_downloader._orchestrator import get_ticker_to_cik_mapping
from sec_edgar_downloader._sec_gateway import download_filing

FileContent = namedtuple("FileContent", ["path", "content"])
DEFAULT_FILTER_PATTERN = "**/*.*"


class Downloader:
    def __init__(
        self,
        company_name: str,
        email_address: str,
    ):
        self.company_name = company_name
        self.email_address = email_address
        self._ticker_to_cik_mapping = get_ticker_to_cik_mapping(self.user_agent)

    @property
    def user_agent(self):
        return f"{self.company_name} {self.email_address}"

    def get_filing_metadatas(
        self,
        query: Union[str, RequestedFilings, CompanyAndAccessionNumber],
    ) -> list[FilingMetadata]:
        if isinstance(query, (CompanyAndAccessionNumber, str)):
            if isinstance(query, str):
                new_query = CompanyAndAccessionNumber.from_string(
                    query, must_match=False
                )
                if new_query is not None:
                    query = new_query
            if isinstance(query, CompanyAndAccessionNumber):
                return [
                    get_filing_metadata(
                        ticker_or_cik=query.ticker_or_cik,
                        accession_number=query.accession_number,
                        user_agent=self.user_agent,
                        ticker_to_cik_mapping=self._ticker_to_cik_mapping,
                    )
                ]

        if isinstance(query, (RequestedFilings, str)):
            if isinstance(query, str):
                query = RequestedFilings.from_string(query)

            new_metadatas = get_latest_filings_metadata(
                requested=query,
                user_agent=self.user_agent,
                ticker_to_cik_mapping=self._ticker_to_cik_mapping,
            )
            return new_metadatas

        raise ValueError(f"Invalid input: {query}")

    def download_filing(self, *, url: str) -> bytes:
        assert url.startswith("https://www.sec.gov/")
        return download_filing(url, self.user_agent)

    def get_filing_html(
        self,
        *,
        query: Optional[Union[str, RequestedFilings, CompanyAndAccessionNumber]] = None,
        # Syntactic Sugar
        ticker: Optional[str] = None,
        form: Optional[str] = None,
    ) -> list[bytes]:
        # Syntactic Sugar
        if query:
            msg = (
                "Error: Ticker or form should not be provided when query is specified."
            )
            assert not ticker and not form, msg
        if ticker or form:
            msg = (
                "Error: Query should not be provided when ticker or form is specified."
            )
            assert not query, msg
            query = f"{ticker}/{form}"
        assert query, "Error: Either query or ticker and form must be specified."

        result = []
        for metadata in self.get_filing_metadatas(query):
            html = self.download_filing(url=metadata.primary_doc_url)
            result.append(html)
        return result
